import argparse
import json
from rich.json import JSON
from rich.prompt import Confirm
from audiodotturn.config import Config
from audiodotturn.errors import error_handler


class Set(Config):
    """
    A class that provides functionality for setting the default config values.
    """
    def __init__(self, args: argparse.Namespace):
        """
        Initialize the Set class object.

        Args:
            args: An argparse.Namespace object containing the command line arguments passed to the Create class object.

        Returns:
            None
        """
        self.args = args
        super().__init__()

    def run(self):
        """
        Sets the default values for the program options using the other Set methods.
        """
        updated = False
        changes = {
            "changed": [],
            "unchanged": []
        }
        self.console.log(f'\nMaking changes in `{self.config_path}`\n')
        for arg_name, arg_value in vars(self.args).items():
            if arg_name not in ['dry', 'command', 'version', 'default'] and arg_value is not None:
                _store_set = None
                if arg_name == "dry_set":
                    _store_set = arg_name
                    arg_name = "dry"
                self.console.log(f"\n[bold magenta]Changes being made to: [bold cyan]{arg_name} -> [bold green]{str(arg_value)} [bold magenta]from [bold blue]{getattr(self, arg_name)}\n")
                try:
                    confirmed = Confirm.ask(prompt="\n        [bold white]Are you sure you want to make these changes?\n", console=self.console)
                except (EOFError, KeyboardInterrupt):
                    self.console.log("\n[bold red]Changes cancelled.\n")
                    self.console.log("[bold red]Exiting.\n")
                    return
                if not confirmed:
                    changes["unchanged"].append(arg_name)
                    self.console.log(f"\n        [yellow]Setting '{arg_name}' will not be changed.\n")
                updated = bool(confirmed)
                if updated:
                    self.console.log(f"\n[bold green] Okay!\n\n        Setting [bold cyan]{arg_name} -> [bold green]{str(arg_value)}\n")
                    if _store_set:
                        arg_name = _store_set
                    changes["changed"].append(arg_name)
                    run = getattr(self, f'set_{arg_name}')
                    run(arg_value)

        self.console.log("[bold green]Short summary of changes: \n", JSON(json.dumps(changes)))
        if changes["changed"]:
            updated = True

        with self.console.status(f"[bold green]Attempting to write to {self.config_path}..."):
            try:
                if updated:
                    self.write_config(dry=self.args.dry)
                    if self.args.dry:
                        self.console.log("\n[green]Defaults would have been updated successfully!")
                    else:
                        self.console.log(f"\n[green]Success! Defaults saved in {self.config_path}!")
                else:
                    self.console.log('\n[green]No changes made!\n')
            except IOError as error:
                self.console.log("\n[red]No settings were updated.\n")
                error_handler(self.error_msg, self.console, error)



    def write_config(self, dry: bool = False) -> None:
        """
        Writes the current configuration data to the JSON file. If dry is True, prints the hypothetical
        changes without actually writing them to the file.

        Args:
            dry: bool flag for dry run
        """
        if dry:
            self.console.log("\n\n[yellow]DRY RUN MODE. NO CHANGES WILL BE WRITTEN TO THE CONFIG FILE.[/yellow]\n\n" + "Hypothetical config created:\n\n" + json.dumps(self.defaults, indent=4))
            return
        try:
            with open(self.config_path, "w", encoding='utf-8') as conf:
                json.dump(self.defaults, conf, indent=4)
        except IOError as error:
            error_handler(self.error_msg, self.console, error)
        
        self.console.log("\n\n[green]CHANGES SAVED TO THE CONFIG FILE.[/green]")

    def set_artist(self, new_artist: str) -> None:
        """
        Sets the default value for artist.

        Args:
            new_artist: string to use as default unknown artist format
        """
        self.defaults["settings"]["formatting_defaults"]["artist"] = new_artist

    def set_title(self, new_title: str) -> None:
        """
        Sets the default value for title.
    
        Args:
            new_title: string to use as default unknown title format
        """
        self.defaults["settings"]["formatting_defaults"]["title"] = new_title

    def set_features(self, new_features: str) -> None:
        """
        Sets the default value for features.

        Args:
            new_features: string to use as default unknown features format
        """
        self.defaults["settings"]["formatting_defaults"]["features"] = new_features

    def set_misc(self, new_misc: str) -> None:
        """
        Sets the default value for misc.

        Args:
            new_misc: string to use as default unknown misc format
        """
        self.defaults["settings"]["formatting_defaults"]["misc"] = new_misc

    def set_youtube_id(self, new_youtube_id: str) -> None:
        """
        Sets the default value for YouTube ID.

        Args:
            new_youtube_id: string to use as default unknown youtube_id format
        """
        self.defaults["settings"]["formatting_defaults"]["youtube_id"] = new_youtube_id

    def set_filetype(self, new_filetype: str) -> None:
        """
        Sets the filetype in the configuration data.

        Args:
            new_filetype: string to use as default unknown filetype format
        """
        self.defaults["settings"]["formatting_defaults"]["filetype"] = new_filetype

    def set_dry_set(self, new_dry_set: bool) -> None:
        """
        Sets the dry flag in the program defaults in the configuration data.

        Args:
            new_dry_set: string with value corresponding to the default dry run status
        """
        self.defaults["settings"]["program_defaults"]["dry"] = new_dry_set

    def set_filename(self, new_filename: str) -> None:
        """
        Sets the filename in the program defaults in the configuration data.

        Args:
            new_filename: string with value corresponding to the default json filename
        """
        self.defaults["settings"]["program_defaults"]["filename"] = new_filename

    def set_directory(self, new_directory: str) -> None:
        """
        Sets the directory in the program defaults in the configuration data.

        Args:
            new_directory: string with value corresponding to the default working directory
        """
        self.defaults["settings"]["program_defaults"]["directory"] = new_directory

    def set_formatter(self, new_formatter: str) -> None:
        """
        Sets the formatter in the program defaults in the configuration data.

        Args:
            new_formatter: string with value corresponding to an avaliable formatter
        """
        self.defaults["settings"]["program_defaults"]["formatter"] = new_formatter

    def set_error_msg(self, new_error_msg: str) -> None:
        """
        Sets the error message in the program defaults in the configuration data.
        
        Args:
            new_error_msg: string formatted as 'option | option | option' with error message formats you'd like audiodotturn to display
        """
        self.defaults["settings"]["program_defaults"]["error_msg"] = new_error_msg

    def set_exts(self, new_exts: str) -> None:
        """
        Sets the extensions in the program defaults in the configuration data.
        
        Args:
            new_exts: string formatted as '.ext1, .ext2, .ext3' with extensions you'd like audiodotturn to format
        """
        self.defaults["settings"]["program_defaults"]["exts"] = new_exts



import argparse
from rich.markdown import Markdown
from audiodotturn.config import defaults
from audiodotturn import VERSION
from audiodotturn.parsers import Parser
from audiodotturn.commands import Create, View, Set

# Get default settings

class CLI:
    """
    A command-line interface runner that handles the parsing of
    arguments and their subsequent execution.
    """
    def __init__(self, args: argparse.Namespace, parsers: tuple((argparse.ArgumentParser, ...))):
        """
        Initializes the CLI with the given arguments and parsers.

        Args:
            args (argparse.Namespace): The arguments passed to the CLI.
            parsers (tuple((argparse.ArgumentParser, ...))): The parsers for the CLI.
        
        Returns:
            None
        """
        self.args = args
        self.parser = parsers[0]
        self.create_parser = parsers[1]
        self.view_parser = parsers[2]

    def run(self) -> None:
        """
        Executes the appropriate action based on the CLI arguments.

        Returns:
            None
        """
        if self.args.version:
            self.show_version()

        elif self.args.defaults:
            self.show_defaults()

        elif self.args.command == 'set':
            setter = Set(self.args)
            setter.run()

        elif self.args.command == 'create':
            creator = Create(self.args)
            # If no args send help text
            if not any((self.args.dirs, self.args.formatfile, self.args.formatdir, self.args.dump)):
                self.create_parser.print_help()
                return
            # run creator
            creator.run()

        elif self.args.command == 'view':
            # initialize viewer
            viewer = View(self.args)
            # If no data arg send help text
            if not self.args.data:
                self.view_parser.print_help()
                return
            # run viewer
            if any(self.args.view_command):
                viewer.run()

        else:
            self.parser.print_help()

    def show_version(self) -> None:
        """
        Shows the current version of the program.

        Returns:
            None
        """
        defaults.console.print(Markdown(f'# {VERSION}'))

    def show_defaults(self):
        """
        Shows the default values for the program options.
    
        Returns:
            None
        """
        display = self.args.defaults
        if display == 'options':
            defaults.display_options()
        elif display == 'format':
            defaults.display_formatting_defaults()
        elif display == 'program':
            defaults.display_program_defaults()
        else:
            defaults.display_all()

def main() -> None:
    """
    Get args from main parser    
    Initialize and start the CLI runner

    Returns:
        None
    """
    # Run cli
    parser = Parser()
    parsers = parser.get_parsers()
    args = parser.parse_args()
    cli = CLI(args=args, parsers=parsers)
    cli.run()

if __name__ == "__main__":
    main()



"""
Module used to initialize CONFIG_PATH for audiodotturn config.

ADT_CONFIG_PATH is a list of common configuration paths for a Linux environment.

.. todo::
   Add option for shell env to be set for ADT_CONFIG_PATH.

.. note::
   Failure of this module will result in sys.exit.

"""
import os
import sys
from rich.console import Console
import pkg_resources

console = Console()
# Check common configuration paths
ADT_CONFIG_PATH = [
    os.path.expanduser("~/.config/audiodotturn/adt_config.json"),
    os.path.expanduser("~/config/audiodotturn/adt_config.json"),
    os.path.expanduser("~/audiodotturn/adt_config.json"),
    os.path.expanduser("~/adt_config.json"),
    "/usr/local/etc/audiodotturn/adt_config.json",
    "/etc/audiodotturn/adt_config.json",
]

CONFIG_PATH = None

for path in ADT_CONFIG_PATH:
    if os.path.exists(path):
        CONFIG_PATH = path
        break

if not CONFIG_PATH:
    console.log('[bold yellow]User config not found, falling back to default config.')
    CONFIG_PATH = pkg_resources.resource_filename(__name__, "config.json")
    if not CONFIG_PATH:
        console.log('[bold red]Error loading default config.')
        sys.exit(1)



import os
import re
import json
import shutil
import argparse
from rich.tree import Tree
from rich.json import JSON
from rich.markdown import Markdown
from audiodotturn.config import Config
from audiodotturn.errors import error_handler


class Create(Config):
    """
    Class that provides functionality for creating formatted files, directories and JSON data.
    """

    def __init__(self, args: argparse.Namespace):
        """
        Initialize the Create class object.

        Args:
            args: An argparse.Namespace object containing the command line arguments passed to the Create class object.

        Returns:
            None
        """
        # grab defaults
        self.args = args
        super().__init__()
        # set user defined values
        self.dry = args.dry
        self.filename = self.args.filename
        self.directory = self.args.directory
        self.formatter = self.args.formatter.lower().strip()
        self.msg = None

    # create runner
    def run(self) -> None:
        """
        Run the Create class.

        Args:
            None

        Returns:
            None
        """
        with self.console.status("[bold green]Working..."):
            if self.args.dirs:
                self.msg = self.dirs()
            elif self.args.formatfile:
                self.filename = self.args.formatfile
                self.msg = f"### **{self.format_file()}**"
            elif self.args.formatdir:
                self.msg = self.format_files_dir()
            elif self.args.dump:
                tree = Tree("DUMP")
                dump = self.json_dump()
                if type(dump) is tuple:
                    tree.add("[green]Runtime Info").add(JSON(dump[0]), highlight=True)
                    tree.add("[cyan]Data").add(JSON(dump[1]), highlight=True)
                else:
                    tree.add("[red]Runtime Info").add(JSON(dump), highlight=True)
                return self.console.log(tree)

            return self.console.log(Markdown(self.msg))

    # check whether a file is formatted, unable to be formatted, or
    # unchanged due to already being formatted or an incompatible extension
    def form_or_not(
        self, files_list: list
    ) -> tuple[list[str], list[str], list[str], list[str]]:
        """
        Check whether a file is formatted, unable to be formatted, or unchanged due to already being formatted or an incompatible extension.

        Args:
            files_list: A list of filenames.

        Returns:
            tuple: A tuple containing four lists: the formatted files, the unformatted files, the already formatted files, and the statistics of the number of each type of file.
        """
        formatted = ["\n## Formatted:"]
        not_formatted = ["\n## Not Formatted:"]
        already_formatted = ["\n## No Change:"]
        for file in files_list:
            if file.startswith("*****"):
                not_formatted.append(f"    - {file.lstrip('*')}")
            elif file.startswith("$$"):
                already_formatted.append(f"    - {file.lstrip('$$')}")
            else:
                formatted.append(f"    - {file}")
        stats = [
            f"\n### Number of files formatted: {len(formatted) - 1}",
            f"\n### Number of files unable to be formatted: {len(not_formatted) - 1}",
            f"\n### Number of files unchanged: {len(already_formatted) - 1}",
        ]

        return formatted, not_formatted, already_formatted, stats

    # create directories based off of unique artists in music library
    def dirs(self) -> str:
        """
        Create directories based off of unique artists in music library.

        Args:
            None

        Returns:
            str: A string message about the directories created and the number of files that were organized.
        """
        directory = self.directory.rstrip("/")
        artists, files = set(), set()
        created = []

        for filename in os.listdir(self.directory):
            if filename.endswith(self.exts):
                match = re.match(
                    r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", filename
                )
                if not match:
                    match = re.match(
                        r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", filename
                    )
                if match:
                    self.artist = match.group(1)
                files.add(self.artist)
                artist_dir = f"{directory}/{self.artist.title()}"
                if not os.path.exists(artist_dir):
                    artists.add(self.artist)
                    if not self.dry:
                        os.makedirs(artist_dir)
                    created.append(f"                    -{artist_dir}/")
                # if dry run, no files will be moved and no directories created,
                # the user just get the data of what wouldve been changed if applied
                if not self.dry:
                    try:
                        shutil.move(
                            f"{self.directory}/{filename}", f"{artist_dir}/{filename}"
                        )
                    except FileExistsError:
                        self.console.log(
                            f"{self.directory}/{filename} unsuccesful move to {artist_dir}/{filename}"
                        )
        return (
            f"### Organized {len(files)} files for {len(artists)} artists.\n"
            + "#### Created directories:\n"
            + "\n".join(created)
        )

    # Attempt to format all files in a given directory, uses form_or_not for listing data to user, dry run settings are handled for this in format_file
    def format_files_dir(self) -> str:
        """
        Goes through all files in the given directory, formats the ones that can be formatted, and returns a string
        summarizing the results of the formatting.

        Returns:
            str: A string summarizing the results of the formatting.
        """
        directory = self.directory.rstrip("/")
        try:
            # go through all files in given directory and only grab the actual files, not directory names and such
            # attempt to format the files as they are found
            files_list = [
                self.format_file(file=f)
                for f in os.listdir(directory)
                if os.path.isfile(directory + "/" + f)
            ]
            formatted, not_formatted, already_formatted, stats = self.form_or_not(
                files_list
            )
            return "\n".join(formatted + not_formatted + already_formatted + stats)
        except (TypeError, FileNotFoundError) as error:
            error_handler(self.error_msg, self.console, error)

    # file argument really only used by format_files_dir, otherwise will use self.filename
    def format_file(self, file: str = None) -> str:
        """
        Formats a single file name according to the formatter specified in the class.

        Args:
            file: A string representing the name of the file to be formatted. If None, uses the filename specified in the class.

        Returns:
            str: A string representing the formatted file name.
        """
        # get filename and create a temp filename
        if file:
            self.filename = file
        file = self.filename

        if self.formatter in ["youtube", "yt"]:
            new_file = self.youtube_formatter(self.filename)
        elif self.formatter in ["standard", "default"]:
            new_file = self.standard_formatter(self.filename)

        # if not a dry-run, set the new file name
        if self.filename != new_file and not self.dry:
            os.rename(
                self.directory + "/" + self.filename, self.directory + "/" + new_file
            )

        return new_file

    def json_dump(self) -> str | tuple[str, str]:
        """
        Finds all properly formatted files in the directory specified in the class and writes the resulting data to a JSON file.
        If dry-run is enabled, does not actually write the file, but instead returns a string representation of the data that
        would have been written to the file. Returns a tuple of two strings in this case.

        Returns:
            str: A string summarizing the result of the JSON dump. If dry-run is enabled, returns a tuple of two strings instead.
        """
        # check if valid directory
        if not os.path.isdir(self.directory):
            return json.dumps(
                {
                    "Result": {
                        "Dry run": f"{self.dry}",
                        "Status": "FAILED",
                        "Final": f"Data has FAILED to be dumped into {self.filename}.",
                        "Reason": f"Directory path '{self.directory}' not found.",
                        "Possible solution": "Ensure the directory exists and its path is correct."
                    }
                }
            )

        # init empty dataset
        data = {}
        for root, _, files in os.walk(self.directory):
            for file in files:
                # Parse filename using regular expressions, if formatted then grab info to add to dataset
                match = re.match(
                    r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", file
                )
                if match:
                    artist, title, features, misc, youtube_id, filetype = match.groups()
                else:
                    match = re.match(
                        r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", file
                    )
                    if match:
                        artist, title, features, misc, filetype = match.groups()
                        youtube_id = None

                if match:
                    # Add track data to nested dictionary
                    if artist not in data:
                        data[artist] = {"tracks": []}

                    if youtube_id:
                        data[artist]["tracks"].append(
                            {
                                "title": title,
                                "features": features,
                                "misc": misc,
                                "youtube_id": youtube_id,
                                "filetype": filetype,
                            }
                        )

                    else:
                        data[artist]["tracks"].append(
                            {
                                "title": title,
                                "features": features,
                                "misc": misc,
                                "filetype": filetype,
                            }
                        )

        if not data:
            return json.dumps(
                {
                    "Result": {
                        "Dry run": f"{self.dry}",
                        "Status": "FAILED",
                        "Final": f"Data has FAILED to be dumped into {self.filename}.",
                        "Reason": "No data available. Most likely cause is having no formatted files available in directory.",
                        "Possible solutions": "Check and make sure you have properly formatted filenames, audiodotturn will ignore unformatted files during the data dump."
                    }
                }
            )

        # if not a dry run then create the json dump in the given location
        # otherwise just print it
        if not self.dry:
            with open(os.path.join(self.filename), "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2)
            return json.dumps(
                {
                    "Result": {
                        "Dry run": f"{self.dry}",
                        "Status": "SUCCESS",
                        "Final": f"Data has been dumped into '{self.filename}'.",
                    }
                }
            )

        # return json.dumps(data, indent=2)
        return json.dumps(
            {
                "Result": {
                    "Dry run": f"{self.dry}",
                    "Status": "SUCCESS",
                    "Final": f"Data would have been dumped into '{self.filename}'.",
                }
            }
        ), json.dumps(data)

    # FORMATTERS
    def youtube_formatter(self, file: str) -> str:
        """
        Formats a given filename string to match the format used by YouTube videos.
        The format is as follows:
        [Artist][Title][Features][Misc][YouTube ID].[Filetype]

        The method checks for existing data in the filename and fills in missing data with defaults
        specified in the configuration file.

        Args:
            file (str): The name of the file to be formatted.

        Returns:
            str: The formatted filename string.
        """
        # check if already formatted or if not an audio file, if so add $$ to temp name
        format_check = re.search(
            r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", file
        )
        if format_check or not file.endswith(self.exts):
            return "$$" + file

        # check for track features
        features_1 = re.search(r"\([fF]t[\. | ](.+?)\)", file)
        if features_1:
            file = file.replace(features_1[0], "")

        features_2 = re.search(
            r"([fF]t[\. | ]|[wW]\/)(.+?)(?=([\'\"\.]|[()]|[-]|[\[]))", file
        )
        if features_2:
            file = file.replace(features_2[0].rstrip("-[("), "")

        features_3 = re.search(r"\([fF]eat[\. | ](.+?)\)", file)
        if features_3:
            file = file.replace(features_3[0], "")

        features_4 = re.search(
            r"([fF]eat[\. | ](.+?)(?=([\'\"\.]|[()]|[-]|[\[])))", file
        )
        if features_4:
            file = file.replace(features_4[0].rstrip("-["), "")

        features = [features_1, features_2, features_3, features_4]

        # check for misc info
        match_misc_1 = re.findall(r"\(.+?\)", file)
        match_misc_2 = re.findall(r"(\[.+?\](?!\.))[ +?]", file)
        match_misc = match_misc_1 + match_misc_2

        if match_misc:
            for match in match_misc:
                file = file.replace(match, "").replace("  ", " ")
                match_misc[match_misc.index(match)] = match.strip("()[] ")

        # check for rest of values, first for a artist-title combo and then just for artist
        match = re.search(
            r"^(.+?)-(.+?) (\[\S+\])?\.(.*$)|^(.+) (\[\S+\])?\.(.*$)", file
        )

        # at this point if file cant be formatted, add ***** to temp name and move on
        if not match:
            file = self.filename
            return f"*****{file}"

        # if file is formattable, check for existing data and fill it in. Use defaults set in config
        # for cases where no info is available.
        else:
            if any(features):
                self.features = []
                if features_1:
                    self.features.append(features_1.group(1).strip())

                if features_2:
                    self.features.append(features_2.group(2).strip())
                if features_3:
                    self.features.append(features_3.group(1).strip())

                if features_4:
                    self.features.append(features_4.group(2).strip())

                self.features = ", ".join(self.features).replace("'", "")

            features = self.features
            misc = ", ".join(match_misc).strip("()") if match_misc else self.misc

            if not match.group(5):
                artist = match.group(1).strip() if match.group(1) else self.artist
                youtube_id = (
                    match.group(3).strip("[]") if match.group(3) else self.youtube_id
                )
                filetype = (
                    match.group(4).strip().rstrip(".")
                    if match.group(4)
                    else self.filetype
                )
            else:
                artist = match.group(5).strip() if match.group(5) else self.artist
                youtube_id = (
                    match.group(6).strip("[]") if match.group(6) else self.youtube_id
                )
                filetype = (
                    match.group(7).strip().rstrip(".")
                    if match.group(7)
                    else self.filetype
                )

            # if there is no title, double check the artist name to see if its possibly located there
            title = match.group(2).strip() if match.group(2) else self.title
            title_in_artist = re.search(r"([\uFF02\"\'].+?[\uFF02\"\'])|(:.+)", artist)
            if title_in_artist and title == "UNKNOWN":
                if title_in_artist.group(1):
                    artist = artist.replace(title_in_artist.group(1), "").strip()
                    title = title_in_artist.group(1)
                else:
                    artist = artist.replace(title_in_artist.group(2), "").strip()
                    title = title_in_artist.group(2)
                title = title.strip("\":\uFF02'")
                title = title.strip()

            # create formatted file name
            return f"[{artist}][{title}][{features}][{misc}][{youtube_id}].{filetype}"

    def standard_formatter(self, file: str) -> str:
        """
        Formats a given filename string to match a standard format.
        The format is as follows:
        [Artist][Title][Features][Misc].[Filetype]

        The method checks for existing data in the filename and fills in missing data with defaults
        specified in the configuration file.

        Args:
            file (str): The name of the file to be formatted.

        Returns:
            str: The formatted filename string.
        """
        # check if already formatted or if not an audio file, if so add $$ to temp name
        format_check = re.search(r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", file)
        format_check_extra = re.search(
            r"\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\[(.+?)\]\.(.+?)$", file
        )
        if format_check or format_check_extra or not file.endswith(self.exts):
            return "$$" + file

        # check for track features

        features_1 = re.search(r"\([fF]t[\. | ](.+?)\)", file)
        if features_1:
            file = file.replace(features_1[0], "")

        features_2 = re.search(
            r"([fF]t[\. | ]|[wW]\/)(.+?)(?=([\'\"\.]|[()]|[-]|[\[]))", file
        )
        if features_2:
            file = file.replace(features_2[0].rstrip("-[("), "")

        features_3 = re.search(r"\([fF]eat[\. | ](.+?)\)", file)
        if features_3:
            file = file.replace(features_3[0], "")

        features_4 = re.search(
            r"([fF]eat[\. | ](.+?)(?=([\'\"\.]|[()]|[-]|[\[])))", file
        )
        if features_4:
            file = file.replace(features_4[0].rstrip("-["), "")

        features = [features_1, features_2, features_3, features_4]

        # check for misc info
        match_misc_search = re.findall(r"(\(.+?\))|(\[.+?\])", file)
        match_misc = []

        if match_misc_search:
            for match in match_misc_search:
                match = match[0] if match[0] else match[1]
                file = file.replace(match, "").replace("  ", " ")
                match_misc.append(match.strip("()[] "))

        # check for rest of values, first for a artist-title combo and then just for artist
        match = re.search(r"^(.+?)-(.+?)\.(.*$)|^(.+)\.(.*$)", file)

        # at this point if file cant be formatted, add ***** to temp name and move on
        if not match:
            file = self.filename
            return f"*****{file}"

        # if file is formattable, check for existing data and fill it in. Use defaults set in config
        # for cases where no info is available.
        else:
            if any(features):
                self.features = []
                if features_1:
                    self.features.append(features_1.group(1).strip())

                if features_2:
                    self.features.append(features_2.group(2).strip())
                if features_3:
                    self.features.append(features_3.group(1).strip())

                if features_4:
                    self.features.append(features_4.group(2).strip())

                self.features = ", ".join(self.features).replace("'", "")

            features = self.features

            misc = ", ".join(match_misc).strip("()") if match_misc else self.misc

            if not match.group(5):
                artist = match.group(1).strip() if match.group(1) else self.artist
                filetype = (
                    match.group(3).strip().rstrip(".")
                    if match.group(3)
                    else self.filetype
                )
            else:
                artist = match.group(4).strip() if match.group(4) else self.artist
                filetype = (
                    match.group(5).strip().rstrip(".")
                    if match.group(5)
                    else self.filetype
                )

            # if there is no title, double check the artist name to see if its possibly located there
            title = match.group(2).strip() if match.group(2) else self.title
            title_in_artist = re.search(r"([\uFF02\"\'].+?[\uFF02\"\'])|(:.+)", artist)
            if title_in_artist and title == "UNKNOWN":
                if title_in_artist.group(1):
                    artist = artist.replace(title_in_artist.group(1), "").strip()
                    title = title_in_artist.group(1)
                else:
                    artist = artist.replace(title_in_artist.group(2), "").strip()
                    title = title_in_artist.group(2)
                title = title.strip("\":\uFF02'")
                title = title.strip()

            # create formatted file name
            return f"[{artist}][{title}][{features}][{misc}].{filetype}"



"""
The `audiodotturn` package contains the source code for AudioDotTurn,
a tool for formatting and organizing audio files with no metadata.

Modules:
- `cli_runner`: defines the command line interface for AudioDotTurn
- `commands`: defines the `create` and `view` subcommands for AudioDotTurn
- `config`: contains the `Config` class, which reads and writes configuration data
- `parsers`: defines parsers for parsing metadata from audio files

Constants:
- `VERSION`: the version of the AudioDotTurn package
"""

__version__ = VERSION = "v0.3.4"



import argparse
from audiodotturn.config import defaults

class Parser:
    """
    A parser to handle command-line arguments for the AudioDotTurn application.
    """
    def __init__(self):
        """
        Initializes the parser with arguments and sub-commands.

        Returns:
            None
        """
        self.parsers = tuple[argparse.ArgumentParser, ...]
        self.parser = argparse.ArgumentParser(description='Format, orgranize and retrieve data from files in an audio library.')
        self.parser.add_argument('--defaults', nargs='?', choices = ['program', 'format', 'options', 'all'], const='all', help='Show default settings')
        self.parser.add_argument('-v', '--version', action='store_true', help='Show current version of audiodotturn')
        self.subparsers = self.parser.add_subparsers(dest='command')
        self.set_parser = self.subparsers.add_parser('set', help='Set defaults')
        self.settings_list = [
            {
                "name": "artist",
                "description": "default artist name for audio files"
            },
            {
                "name": "title",
                "description": "default title for audio files"
            },
            {
                "name": "features",
                "description": "default features for audio files"
            },
            {
                "name": "misc",
                "description": "default miscellaneous info for audio files"
            },
            {
                "name": "youtube_id",
                "description": "default YouTube ID for audio files"
            },
            {
                "name": "filetype",
                "description": "default audio file type"
            },
            {
                "name": "dry_set",
                "description": "default dry run setting"
            },
            {
                "name": "filename",
                "description": "default json filename"
            },
            {
                "name": "directory",
                "description": "default working directory"
            },
            {
                "name": "formatter",
                "description": "default formatter"
            },
            {
                "name": "error_msg",
                "description": "default error message"
            },
            {
                "name": "exts",
                "description": "default formattable file extensions"
            }
        ]

        for setting in self.settings_list:
            self.set_parser.add_argument(f"--{setting['name']}", dest=setting['name'], type=str, help=setting['description'])
        self.set_parser.add_argument("--dry", "--dry_run", action='store_true', help='Dry run')

        self.create_parser = self.subparsers.add_parser('create', help='Create subcommands')
        self.create_parser.add_argument('-d', '--dirs', action='store_true', help='Organize files in artist directories')
        self.create_parser.add_argument('-f', '--formatfile', type=str, help='Format single file')
        self.create_parser.add_argument('-x', '--formatter', type=str, default=defaults.formatter, help='Define the formatter to use.')
        self.create_parser.add_argument('-F', '--formatdir', action='store_true', help='Format all files in directory')
        self.create_parser.add_argument('-D', '--dump', action='store_true', help='Dump directory into JSON file')
        self.create_parser.add_argument('--filename', type=str, default=defaults.filename, help='Name of JSON file')
        self.create_parser.add_argument('--dry', action='store_true', help='Dry run')
        self.create_parser.add_argument('--directory', type=str, default=defaults.directory, help='Directory to organize or format files')

        self.view_parser = self.subparsers.add_parser('view', help='View subcommands')
        self.view_parser.add_argument('-d', '--data', type=str, help='JSON data to view')
        self.view_subparsers = self.view_parser.add_subparsers(dest='view_command')

        # Create parser for the "artists" subcommand of the "view" command
        self.artists_parser = self.view_subparsers.add_parser('artists', help='View list of artists')
        self.artists_parser.add_argument('-t', '--tracks', action='store_true', help='View list of artists and their tracks')
        self.artists_parser.add_argument('-n', '--names', action='store_true', help='View list of artist names')

        self.songs_parser = self.view_subparsers.add_parser('songs', help='View list of songs')
        self.songs_parser.add_argument('-a', '--artist', type=str, help='View list of songs by artist')
        self.songs_parser.add_argument('-i', '--id', type=str, help='View list of songs by ID')
        self.songs_parser.add_argument('-N', '--name', type=str, help='View list of songs by name')


    def get_parsers(self):
        """
        Return a tuple of parsers that can be used to parse command-line arguments.

        Returns:
            tuple: A tuple of parsers that can be used to parse command-line arguments.
        """
        self.parsers = (self.parser, self.create_parser, self.view_parser)
        return self.parsers
    
    def parse_args(self, args=None):
        """
        Parse command-line arguments using the parser.

        Args:
            args (list): The arguments to parse. If None, command-line arguments are used.

        Returns:
            Namespace: argparse.Namespace object of the parsed arguments.
        """
        return self.parser.parse_args(args)



import os
import json
import argparse
from rich.markdown import Markdown
from audiodotturn.config import Config
from audiodotturn.errors import error_handler


class View(Config):
    """
    A class that provides functionality for viewing the contents of the dataset.
    """
    def __init__(self, args: argparse.Namespace):
        """
        Initialize View class

        Args:
            args: An argparse.Namespace object containing the command line arguments passed to the Create class object.

        Returns:
            None
        """
        # get defaults
        self.args = args
        super().__init__()
        self.msg = ''
        # set user defined args
        if self.args.view_command == 'songs':
            self.artist = self.args.artist if self.args.artist else ''
            self.id = self.args.id if self.args.id else ''
            self.name = self.args.name if self.args.name else ''
        # attempt to open the given dataset
        try:
            with open(os.path.join(self.args.data), 'r', encoding='utf-8') as f:
                self.dataset = json.load(f, object_hook=dict)
        except IOError as error:
            self.dataset = {}
            error_handler(self.error_msg, self.console, error)

    # view command runner
    def run(self) -> None:
        """
        Run the View class.

        Args:
            None

        Returns:
            None
        """
        with self.console.status("[bold green]Working..."):
            if self.args.view_command == 'artists':
                if self.args.tracks:
                    self.msg = self.get_artists_tracks()
                elif self.args.names:
                    self.msg = self.get_artists()

            elif self.args.view_command == 'songs':
                if self.args.artist:
                    self.artist = self.args.artist
                    self.msg = self.get_songs_by_artist()
                elif self.args.id:
                    self.id = self.args.id
                    self.msg = self.get_songs_by_id()
                elif self.args.name:
                    self.name = self.args.name
                    self.msg = self.get_songs_by_name()
            else:
                self.msg = self.args.view_command

            self.console.log(Markdown(self.msg))

    # View - artists commands
    def get_artists(self) -> str:
        """
        Returns:
            str: a formatted string containing a list of all artists in the dataset.
        """        
        # get list of all artists in dataset
        artists_list = []
        for artist in self.dataset:
            artists_list.append(f'                            - {artist}')

        return f"# Database: {self.filename}\n\n" + "## Artists:\n" + '\n'.join(artists_list)

    def get_artists_tracks(self) -> str:
        """
        Returns:
            str: a formatted string containing a list of all the artists and their songs in the dataset.
        """
        # get list of all artists and their tracks in dataset
        artists_list = []
        for artist in self.dataset:
            artists_list.append(f'### Artist: **{artist}**')
            for tracks in self.dataset[artist]['tracks']:
                msg = f'''#### Track: {tracks["title"]}
                            - Artist: {artist}
                            - Features: {tracks["features"]}
                            - Misc: {tracks["misc"]}
                            - YouTube ID: {tracks["youtube_id"]}
                            - File Type: {tracks["filetype"]}\n'''
                artists_list.append(msg)

        return f"# Database: {self.filename}\n\n" + "## Artists and their Tracks:\n" + '\n'.join(artists_list)

    # View - songs commands
    def get_songs_by_artist(self) -> str:
        """
        Returns:
            str: a formatted string containing a list of all songs by a specified artist in the dataset.
        """
        # get list of all songs by an artist (will search by substring) name in dataset
        artists_list = []
        for artist_key in self.dataset.keys():
            if self.artist.lower() in artist_key.lower():
                for track in self.dataset[artist_key]['tracks']:
                    msg = f'''#### **{track["title"]}**
                                - Artist: {artist_key}
                                - Features: {track["features"]}
                                - Misc: {track["misc"]}
                                - YouTube ID: {track["youtube_id"]}
                                - File Type: {track["filetype"]}\n'''
                    artists_list.append(msg)

        return f"# Database: {self.filename}\n\n" + f"## Songs by artist matching '{self.artist}':\n" + '\n'.join(artists_list)

    def get_songs_by_id(self) -> str:
        """
        Returns:
            str: a formatted string containing a list of all songs by a specified youtube_id in the 
        dataset.
        """
        # get list of all songs by their youtube id in dataset
        artists_list = []
        for artist_key in self.dataset.keys():
            for track in self.dataset[artist_key]['tracks']:
                if self.youtube_id in track['youtube_id']:
                    msg = f'''#### **{track["title"]}**
                                - Artist: {artist_key}
                                - Features: {track["features"]}
                                - Misc: {track["misc"]}
                                - YouTube ID: {track["youtube_id"]}
                                - File Type: {track["filetype"]}\n'''
                    artists_list.append(msg)

        return f"# Database: {self.filename}\n\n" + f"## Songs by ID {self.youtube_id}:\n" + '\n'.join(artists_list)

    def get_songs_by_name(self) -> str:
        """
        Returns:
            str: a formatted string containing a list of all songs by a specified title/name in the 
        dataset.
        """
        artists_list = []
        for artist_key in self.dataset.keys():
            for track in self.dataset[artist_key]['tracks']:
                if self.name.lower() in track['title'].lower():
                    msg = f'''#### **{track["title"]}**
                                - Artist: {artist_key}
                                - Features: {track["features"]}
                                - Misc: {track["misc"]}
                                - YouTube ID: {track["youtube_id"]}
                                - File Type: {track["filetype"]}\n'''
                    artists_list.append(msg)

        return f"# Database: {self.filename}\n\n" + f"## Tracks with name matching '{self.name}':\n" + "\n".join(artists_list)



import json
from rich.console import Console
from rich.table import Table
from audiodotturn.config.config_path import CONFIG_PATH
from audiodotturn.errors import error_handler

class Config:
    """
    Class for managing configuration data.
    """
    def __init__(self):
        """
        Constructor for Config class. 
        Loads configuration data from JSON file and sets default values.

        Returns:
            None
        """
        self.console = Console()

        # read data from json and set defaults
        try:
            with open(CONFIG_PATH, "r", encoding='utf-8') as c:
                self.defaults = json.load(c)
                self.options = self.defaults["options"]
                self.settings = self.defaults["settings"]

                self.artist = self.settings["formatting_defaults"]["artist"]
                self.title = self.settings["formatting_defaults"]["title"]
                self.features = self.settings["formatting_defaults"]["features"]
                self.misc = self.settings["formatting_defaults"]["misc"]
                self.youtube_id = self.settings["formatting_defaults"]["youtube_id"]
                self.filetype = self.settings["formatting_defaults"]["filetype"]

                self.dry = self.settings["program_defaults"]["dry"].lower()
                self.dry = True if self.dry == "true" else False

                self.filename = self.settings["program_defaults"]["filename"]
                self.directory = self.settings["program_defaults"]["directory"]
                self.formatter = self.settings["program_defaults"]["formatter"]
                self.error_msg = self.settings["program_defaults"]["error_msg"].replace(' ', '').split('|')
                self.exts = tuple(self.settings["program_defaults"]["exts"].replace(' ', '').split(","))
                self.config_path = CONFIG_PATH
        except KeyError as error:
            self.console.log('\n[bold red]Config file not configured correctly. Make sure all values are set, refer to the default config or the docs for examples.')
            error_handler(self.error_msg, self.console, error)

    def display_all(self) -> None:
        """
        Displays all configuration data in a formatted table.

        Returns:
            None
        """
        table = Table(title="Configurations", style="bold magenta")
        table.add_column("Section", justify="center", style="cyan", min_width=30)
        table.add_column("Config", justify="center", style="cyan", min_width=20)
        for key in self.defaults:
            if key == "settings":
                subtable = Table(style="blue")
                subtable.add_column("Setting", justify="center", style="magenta", min_width=30)
                subtable.add_column("Value", justify="center", style="magenta", min_width=20)
                for subkey, subval in self.settings.items():
                    if isinstance(subval, dict):
                        for subsubkey, subsubval in subval.items():
                            subtable.add_row(f"{subkey} - {subsubkey}", str(subsubval))
                    else:
                        subtable.add_row(subkey, str(subval))
                table.add_row("Current Defaults", subtable)
            elif key == "options":
                subtable = Table(style="green")
                subtable.add_column("Option", justify="center", style="yellow", min_width=30)
                subtable.add_column("Value", justify="center", style="yellow", min_width=20)
                for subkey, subval in self.options.items():
                    if isinstance(subval, dict):
                        subsubtable = Table(style="blue")
                        subsubtable.add_column('Description', justify="center", style="yellow", min_width=40)
                        subsubtable.add_column('Options', justify="center", style="yellow", min_width=60)
                        for subsubkey, subsubval in subval.items():
                            if isinstance(subsubval, dict):
                                for sub3key, sub3val in subsubval.items():
                                    if isinstance(sub3val, dict):
                                        opts = ''
                                        for key in sub3val:
                                            opts += f'{key} - {sub3val[key]}\n'
                                        sub3val = opts
                                    subsubtable.add_row(f"{subsubkey} - {sub3key}", str(sub3val) + '\n')
                            else:    
                                subsubtable.add_row(f"{str(subkey).title().replace('_', ' ')} - {subsubkey}", str(subsubval) + '\n')

                        table.add_row(f"{str(subkey).title().replace('_', ' ')} Options", subsubtable)

                    else:
                        subtable.add_row(str(subkey).title().replace('_', ' '), str(subval) + '\n')
        self.console.log(table)

    # Display formatting defaults in a formatted table
    def display_formatting_defaults(self) -> None:
        """
        Displays the formatting defaults in a formatted table.

        Returns:
            None
        """
        table = Table(title="Formatting Defaults", style="bold blue")
        table.add_column("Setting", style="magenta", min_width=10)
        table.add_column("Value", style="magenta", min_width=10)

        for key in self.settings["formatting_defaults"]:
            table.add_row(key, self.settings["formatting_defaults"][key])
        self.console.log(table)

    # Display program defaults in a formatted table
    def display_program_defaults(self) -> None:
        """
        Displays the current defaults in a formatted table.

        Returns:
            None
        """
        table = Table(title="Program Defaults", style="bold green")
        table.add_column("Setting", style="yellow", min_width=10)
        table.add_column("Value", style="yellow", min_width=10)

        for key in self.settings["program_defaults"]:
            table.add_row(key, self.settings["program_defaults"][key])
        self.console.log(table)

    # Display options in a formatted table
    def display_options(self) -> None:
        """
        Displays the config options in a formatted table.

        Returns:
            None
        """
        table = Table(title="Options", show_header=True, header_style="bold magenta")
        table.add_column("Category", style="cyan", justify="center", min_width=30)
        table.add_column("Info", style="magenta", justify="center", min_width=35)
        for subkey, subval in self.options.items():
            if isinstance(subval, dict):
                subsubtable = Table(style="blue")
                subsubtable.add_column('Description', justify="center", style="yellow", min_width=40)
                subsubtable.add_column('Options', justify="center", style="yellow", min_width=60)
                for subsubkey, subsubval in subval.items():
                    if isinstance(subsubval, dict):
                        for sub3key, sub3val in subsubval.items():
                            if isinstance(sub3val, dict):
                                opts = ''
                                for key in sub3val:
                                    opts += f'{key} - {sub3val[key]}\n'
                                sub3val = opts
                            subsubtable.add_row(f"{subsubkey} - {sub3key}", str(sub3val) + '\n')
                    else:    
                        subsubtable.add_row(f"{str(subkey).title().replace('_', ' ')} - {subsubkey}", str(subsubval) + '\n')

                table.add_row(f"{str(subkey).title().replace('_', ' ')} Options", subsubtable)

        self.console.print(table)

#init config
defaults = Config()



import sys
import traceback
from rich.console import Console

def error_handler(error_msg: list, console: Console, error: Exception) -> None:
    """
    General error message handling for audiodotturn

    Args:
        list: error_msg list produced by config, typically self.error_msg
        rich.console.Console: rich console to use
        Exception: Exception being handled

    Returns:
        None: Calls sys.exit()
    """
    if "error" in error_msg:
        console.log(f'[yellow]Error -> [red]{error}')
        error_msg.remove("error")
    if "traceback" in error_msg:
        error_msg.remove("traceback")
        tb = sys.exception().__traceback__
        tb_txt = ''.join(traceback.format_list(traceback.extract_tb(tb)))
        console.log(tb_txt)
    if error_msg:
        console.log('\n'.join(error_msg))
    sys.exit(1)